import re
from typing import Iterable


class CollectionBase:

    def __dir__(self) -> Iterable[str]:
        internal_methods = ['create', 'drop', 'get', 'list']

        items = [item.name for item in self.list()]

        items = [i for i in items if re.match('^(?![0-9])\w+$', i)]
        return internal_methods + items

    def __getattr__(self, name):
        if name.startswith('__'):
            raise AttributeError(name)

        return self.get(name)


# class MethodCollection(CollectionBase):
#
#     def __init__(self, name, methods):
#         self.name = name
#         self.methods = methods
#
#     def __repr__(self):
#         return f'{self.__class__.__name__}({self.name})'
#
#     def get(self, *args, **kwargs):
#         method = self.methods.get('get')
#         if method is None:
#             raise NotImplementedError()
#
#         return method(*args, **kwargs)
#
#     def list(self, *args, **kwargs):
#         method = self.methods.get('list')
#         if method is None:
#             raise NotImplementedError()
#
#         return method(*args, **kwargs)
#
#     def create(self, *args, **kwargs):
#         method = self.methods.get('create')
#         if method is None:
#             raise NotImplementedError()
#
#         return method(*args, **kwargs)
#
#     def drop(self, name):
#         method = self.methods.get('drop')
#         if method is None:
#             raise NotImplementedError()
#
#         return method(name)



import copy
import json
from typing import Union, List

import pandas as pd

from mindsdb_sql.parser.dialects.mindsdb import CreateKnowledgeBase, DropKnowledgeBase
from mindsdb_sql.parser.ast import Identifier, Star, Select, BinaryOperation, Constant, Insert

from mindsdb_sdk.utils.sql import dict_to_binary_op, query_to_native_query
from mindsdb_sdk.utils.objects_collection import CollectionBase
from mindsdb_sdk.utils.context import is_saving

from .models import Model
from .tables import Table
from .query import Query
from .databases import Database


class KnowledgeBase(Query):
    """

    Knowledge base object, used to update or query knowledge base

    Add data to knowledge base:

    >>> kb.insert(pd.read_csv('house_sales.csv'))

    Query relevant results

    >>> df = kb.find('flats').fetch()

    """

    def __init__(self, api, project, data: dict):
        self.api = api
        self.project = project
        self.name = data['name']
        self.table_name = Identifier(parts=[self.project.name, self.name])

        self.storage = None
        if data['storage'] is not None:
            # if name contents '.' there could be errors

            parts = data['storage'].split('.')
            if len(parts) == 2:
                database_name, table_name = parts
                database = Database(project, database_name)
                table = Table(database, table_name)
                self.storage = table

        self.model = None
        if data['model'] is not None:
            self.model = Model(self.project, {'name': data['model']})

        params = data.get('params', {})
        if isinstance(params, str):
            try:
                params = json.loads(params)
            except json.JSONDecodeError:
                params = {}

        # columns
        self.metadata_columns = params.pop('metadata_columns', [])
        self.content_columns = params.pop('content_columns', [])
        self.id_column = params.pop('id_column', None)

        self.params = params

        # query behavior
        self._query = None
        self._limit = None

        self._update_query()

        # empty database
        super().__init__(project.api, self.sql, None)

    def __repr__(self):
        return f'{self.__class__.__name__}({self.project.name}.{self.name})'

    def find(self, query: str, limit: int = 100):
        """

        Query data from knowledge base.
        Knowledge base should return a most relevant results for the query

        >>> # query knowledge base
        >>> query = my_kb.find('dogs')
        >>> # fetch dataframe to client
        >>> print(query.fetch())

        :param query: text query
        :param limit: count of rows in result, default 100
        :return: Query object
        """

        kb = copy.deepcopy(self)
        kb._query = query
        kb._limit = limit
        kb._update_query()

        return kb

    def _update_query(self):

        ast_query = Select(
            targets=[Star()],
            from_table=self.table_name
        )
        if self._query is not None:
            ast_query.where = BinaryOperation(op='=', args=[
                Identifier('content'),
                Constant(self._query)
            ])

        if self._limit is not None:
            ast_query.limit = Constant(self._limit)
        self.sql = ast_query.to_string()

    def insert_files(self, file_paths: List[str]):
        """
        Insert data from file to knowledge base
        """
        self.api.insert_files_into_knowledge_base(self.project.name, self.name, file_paths)

    def insert_webpages(self, urls: List[str], crawl_depth: int = 1, filters: List[str] = None):
        """
        Insert data from crawled URLs to knowledge base.

        :param urls: URLs to be crawled and inserted.
        :param crawl_depth: How deep to crawl from each base URL. 0 = scrape given URLs only
        :param filters: Include only URLs that match these regex patterns
        """
        self.api.insert_webpages_into_knowledge_base(self.project.name, self.name, urls, crawl_depth=crawl_depth, filters=filters)

    def insert(self, data: Union[pd.DataFrame, Query, dict]):
        """
        Insert data to knowledge base

        >>> # insert using query
        >>> my_kb.insert(server.databases.example_db.tables.houses_sales.filter(type='house'))
        >>> # using dataframe
        >>> my_kb.insert(pd.read_csv('house_sales.csv'))
        >>> # using dict
        >>> my_kb.insert({'type': 'house', 'date': '2020-02-02'})

        Data will be if id (defined by id_column param, see create knowledge base) is already exists in knowledge base
        it will be replaced

        :param data: Dataframe or Query object or dict.
        """

        if isinstance(data, dict):
            data = pd.DataFrame([data])

        if isinstance(data, pd.DataFrame):
            # insert data
            data_split = data.to_dict('split')

            ast_query = Insert(
                table=Identifier(self.table_name),
                columns=data_split['columns'],
                values=data_split['data']
            )
            sql = ast_query.to_string()

        else:
            # insert from select
            if data.database is not None:
                ast_query = Insert(
                    table=Identifier(self.table_name),
                    from_select=query_to_native_query(data)
                )
                sql = ast_query.to_string()
            else:
                sql = f'INSERT INTO {self.table_name.to_string()} ({data.sql})'

        if is_saving():
            # don't execute it right now, return query object
            return Query(self, sql, self.database)

        self.api.sql_query(sql, self.database)


class KnowledgeBases(CollectionBase):
    """
    **Knowledge bases**

    Get list:

    >>> kb_list = server.knowledge_bases.list()
    >>> kb = kb_list[0]

    Get by name:

    >>> kb = server.knowledge_bases.get('my_kb')
    >>> # or :
    >>> kb = server.knowledge_bases.my_kb

    Create:

    >>> kb = server.knowledge_bases.create('my_kb')

    Drop:

    >>> server.knowledge_bases.drop('my_kb')

    """

    def __init__(self, project, api):
        self.project = project
        self.api = api

    def _list(self, name: str = None) -> List[KnowledgeBase]:

        # TODO add filter by project. for now 'project' is empty
        ast_query = Select(targets=[Star()], from_table=Identifier(parts=['information_schema', 'knowledge_bases']))
        if name is not None:
            ast_query.where = dict_to_binary_op({'name': name})

        df = self.api.sql_query(ast_query.to_string(), database=self.project.name)

        # columns to lower case
        cols_map = {i: i.lower() for i in df.columns}
        df = df.rename(columns=cols_map)

        return [
            KnowledgeBase(self.api, self.project, item)
            for item in df.to_dict('records')
        ]

    def list(self) -> List[KnowledgeBase]:
        """

        Get list of knowledge bases inside of project:

        >>> kb_list = project.knowledge_bases.list()

        :return: list of knowledge bases
        """
        return self._list()

    def get(self, name: str) -> KnowledgeBase:
        """
        Get knowledge base by name

        :param name: name of the knowledge base
        :return: KnowledgeBase object
        """
        item = self._list(name)
        if len(item) == 1:
            return item[0]
        elif len(item) == 0:
            raise AttributeError("KnowledgeBase doesn't exist")
        else:
            raise RuntimeError("Several knowledgeBases with the same name")

    def create(
        self,
        name: str,
        model: Model = None,
        storage: Table = None,
        metadata_columns: list = None,
        content_columns: list = None,
        id_column: str = None,
        params: dict = None,
    ) -> Union[KnowledgeBase, Query]:
        """

        Create knowledge base:

        >>> kb = server.knowledge_bases.create(
        ...   'my_kb',
        ...   model=server.models.emb_model,
        ...   storage=server.databases.pvec.tables.tbl1,
        ...   metadata_columns=['date', 'author'],
        ...   content_columns=['review', 'description'],
        ...   id_column='number',
        ...   params={'a': 1}
        ...)

        :param name: name of the knowledge base
        :param model: embedding model, optional. Default: 'sentence_transformers' will be used (defined in mindsdb server)
        :param storage: vector storage, optional. Default: chromadb database will be created
        :param metadata_columns: columns to use as metadata, optional. Default: all columns which are not content and id
        :param content_columns: columns to use as content, optional. Default: all columns except id column
        :param id_column: the column to use as id, optinal. Default: 'id', if exists
        :param params: other parameters to knowledge base
        :return: created KnowledgeBase object
        """

        params_out = {}

        if metadata_columns is not None:
            params_out['metadata_columns'] = metadata_columns

        if content_columns is not None:
            params_out['content_columns'] = content_columns

        if id_column is not None:
            params_out['id_column'] = id_column

        if params is not None:
            params_out.update(params)

        if model is not None:
            model_name = Identifier(parts=[model.project.name, model.name])
        else:
            model_name = None

        if storage is not None:
            storage_name = Identifier(parts=[storage.db.name, storage.name])
        else:
            storage_name = None

        ast_query = CreateKnowledgeBase(
            Identifier(parts=[self.project.name, name]),
            model=model_name,
            storage=storage_name,
            params=params_out
        )
        sql = ast_query.to_string()

        if is_saving():
            return Query(self, sql)

        self.api.sql_query(sql)

        return self.get(name)

    def drop(self, name: str):
        """

        :param name:
        :return:
        """

        ast_query = DropKnowledgeBase(Identifier(parts=[self.project.name, name]))
        sql = ast_query.to_string()

        if is_saving():
            return Query(self, sql)

        self.api.sql_query(sql)



from requests.exceptions import HTTPError
from typing import Iterable, List, Union
from urllib.parse import urlparse
from uuid import uuid4
import datetime
import json

from mindsdb_sdk.knowledge_bases import KnowledgeBase
from mindsdb_sdk.models import Model
from mindsdb_sdk.skills import Skill
from mindsdb_sdk.utils.objects_collection import CollectionBase

_DEFAULT_LLM_MODEL = 'gpt-4o'
_DEFAULT_LLM_PROMPT = 'Answer the user"s question in a helpful way: {{question}}'


class AgentCompletion:
    """
    Represents a full MindsDB agent completion response.

    Attributes:
    content: The completion content.
    context: Only relevant for retrieval agents. Contains the context retrieved from the knowledge base.


    """

    def __init__(self, content: str, context: List[dict] = None):
        self.content = content
        self.context = context

    def __repr__(self):
        return f'{self.__class__.__name__}(content: {self.content}, context: {self.context})'


class Agent:
    """Represents a MindsDB agent.
    
    Working with agents:

    Get an agent by name:

    >>> agent = agents.get('my_agent')

    Query an agent:

    >>> completion = agent.completion([{'question': 'What is your name?', 'answer': None}])
    >>> print(completion.content)

    Query an agent with streaming:

    >>> completion = agent.completion_stream([{'question': 'What is your name?', 'answer': None}])
    >>> for chunk in completion:
            print(chunk.choices[0].delta.content)

    List all agents:

    >>> agents = agents.list()

    Create a new agent:

    >>> model = models.get('my_model') # Or use models.create(...)
    >>> # Connect your agent to a MindsDB table.
    >>> text_to_sql_skill = skills.create('text_to_sql', 'sql', { 'tables': ['my_table'], 'database': 'my_database' })
    >>> agent = agents.create('my_agent', model, [text_to_sql_skill])

    Update an agent:

    >>> new_model = models.get('new_model')
    >>> agent.model_name = new_model.name
    >>> new_skill = skills.create('new_skill', 'sql', { 'tables': ['new_table'], 'database': 'new_database' })
    >>> updated_agent.skills.append(new_skill)
    >>> updated_agent = agents.update('my_agent', agent)

    Delete an agent by name:

    >>> agents.drop('my_agent')
    """

    def __init__(
            self,
            name: str,
            model_name: str,
            skills: List[Skill],
            params: dict,
            created_at: datetime.datetime,
            updated_at: datetime.datetime,
            provider: str = None,
            collection: CollectionBase = None
    ):
        self.name = name
        self.model_name = model_name
        self.provider = provider
        self.skills = skills
        self.params = params
        self.created_at = created_at
        self.updated_at = updated_at
        self.collection = collection

    def completion(self, messages: List[dict]) -> AgentCompletion:
        return self.collection.completion(self.name, messages)

    def completion_stream(self, messages: List[dict]) -> Iterable[object]:
        return self.collection.completion_stream(self.name, messages)

    def add_files(self, file_paths: List[str], description: str, knowledge_base: str = None):
        """
        Add a list of files to the agent for retrieval.

        :param file_paths: List of paths to the files to be added.
        """
        self.collection.add_files(self.name, file_paths, description, knowledge_base)

    def add_file(self, file_path: str, description: str, knowledge_base: str = None):
        """
        Add a file to the agent for retrieval.

        :param file_path: Path to the file to be added.
        """
        self.collection.add_file(self.name, file_path, description, knowledge_base)

    def add_webpages(
            self,
            urls: List[str],
            description: str,
            knowledge_base: str = None,
            crawl_depth: int = 1,
            filters: List[str] = None):
        """
        Add a crawled URL to the agent for retrieval.

        :param urls: URLs of pages to be crawled and added.
        :param description: Description of the webpages. Used by agent to know when to do retrieval.
        :param knowledge_base: Name of an existing knowledge base to be used. Will create a default knowledge base if not given.
        :param crawl_depth: How deep to crawl from each base URL. 0 = scrape given URLs only, -1 = default max
        :param filters: Include only URLs that match these regex patterns
        """
        self.collection.add_webpages(self.name, urls, description, knowledge_base=knowledge_base, crawl_depth=crawl_depth, filters=filters)

    def add_webpage(
            self,
            url: str,
            description: str,
            knowledge_base: str = None,
            crawl_depth: int = 1,
            filters: List[str] = None):
        """
        Add a crawled URL to the agent for retrieval.

        :param url: URL of the page to be crawled and added.
        :param description: Description of the webpages. Used by agent to know when to do retrieval.
        :param knowledge_base: Name of an existing knowledge base to be used. Will create a default knowledge base if not given.
        :param crawl_depth: How deep to crawl from each base URL. 0 = scrape given URLs only, -1 = default max
        :param filters: Include only URLs that match these regex patterns
        """
        self.collection.add_webpage(self.name, url, description, knowledge_base=knowledge_base, crawl_depth=crawl_depth, filters=filters)

    def add_database(self, database: str, tables: List[str], description: str):
        """
        Add a database to the agent for retrieval.

        :param database: Name of the database to be added.
        :param tables: List of tables to be added.
        :param description: Description of the database tables. Used by the agent to know when to use SQL skill.
        """
        self.collection.add_database(self.name, database, tables, description)

    def __repr__(self):
        return f'{self.__class__.__name__}(name: {self.name})'

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.model_name != other.model_name:
            return False
        if self.provider != other.provider:
            return False
        if self.skills != other.skills:
            return False
        if self.params != other.params:
            return False
        if self.created_at != other.created_at:
            return False
        return self.updated_at == other.updated_at

    @classmethod
    def from_json(cls, json: dict, collection: CollectionBase):
        return cls(
            json['name'],
            json['model_name'],
            [Skill.from_json(skill) for skill in json['skills']],
            json['params'],
            json['created_at'],
            json['updated_at'],
            json['provider'],
            collection
        )


class Agents(CollectionBase):
    """Collection for agents"""

    def __init__(self, project, api):
        self.api = api
        self.project = project

        self.knowledge_bases = project.knowledge_bases
        self.models = project.models
        self.skills = project.skills

        self.databases = project.server.databases
        self.ml_engines = project.server.ml_engines

    def list(self) -> List[Agent]:
        """
        List available agents.

        :return: list of agents
        """
        data = self.api.agents(self.project.name)
        return [Agent.from_json(agent, self) for agent in data]

    def get(self, name: str) -> Agent:
        """
        Gets an agent by name.

        :param name: Name of the agent

        :return: agent with given name
        """
        data = self.api.agent(self.project.name, name)
        return Agent.from_json(data, self)

    def completion(self, name: str, messages: List[dict]) -> AgentCompletion:
        """
        Queries the agent for a completion.

        :param name: Name of the agent
        :param messages: List of messages to be sent to the agent

        :return: completion from querying the agent
        """
        data = self.api.agent_completion(self.project.name, name, messages)
        if 'context' in data['message']:
            return AgentCompletion(data['message']['content'], data['message'].get('context'))

        return AgentCompletion(data['message']['content'])

    def completion_stream(self, name, messages: List[dict]) -> Iterable[object]:
        """
        Queries the agent for a completion and streams the response as an iterable object.

        :param name: Name of the agent
        :param messageS: List of messages to be sent to the agent

        :return: iterable of completion chunks from querying the agent.
        """
        return self.api.agent_completion_stream(self.project.name, name, messages)

    def _create_default_knowledge_base(self, agent: Agent, name: str) -> KnowledgeBase:
        # Make sure default ML engine for embeddings exists.
        try:
            _ = self.ml_engines.get('langchain_embedding')
        except AttributeError:
            _ = self.ml_engines.create('langchain_embedding', 'langchain_embedding')
        # Include API keys in embeddings.
        if agent.provider == "mindsdb":
            agent_model = self.models.get(agent.model_name)
            training_options = json.loads(agent_model.data.get('training_options', '{}'))
            training_options_using = training_options.get('using', {})
            api_key_params = {k: v for k, v in training_options_using.items() if 'api_key' in k}
            kb = self.knowledge_bases.create(name, params=api_key_params)
        else:
            kb = self.knowledge_bases.create(name)
        # Wait for underlying embedding model to finish training.
        kb.model.wait_complete()
        return kb

    def add_files(self, name: str, file_paths: List[str], description: str, knowledge_base: str = None):
        """
        Add a list of files to the agent for retrieval.

        :param name: Name of the agent
        :param file_paths: List of paths or URLs to the files to be added.
        :param description: Description of the file. Used by agent to know when to do retrieval
        :param knowledge_base: Name of an existing knowledge base to be used. Will create a default knowledge base if not given.
        """
        if not file_paths:
            return
        filename_no_extension = ''
        all_filenames = []
        for file_path in file_paths:
            filename = file_path.split('/')[-1]
            filename_no_extension = filename.split('.')[0]
            all_filenames.append(filename_no_extension)
            try:
                _ = self.api.get_file_metadata(filename_no_extension)
            except HTTPError as e:
                if e.response.status_code >= 400 and e.response.status_code != 404:
                    raise e
                # upload file to mindsdb
                self.api.upload_file(filename, file_path)

        # Insert uploaded files into new knowledge base.
        agent = self.get(name)
        if knowledge_base is not None:
            kb = self.knowledge_bases.get(knowledge_base)
        else:
            kb_name = f'{name}_{filename_no_extension}_{uuid4()}_kb'
            kb = self._create_default_knowledge_base(agent, kb_name)

        # Insert the entire file.
        kb.insert_files(all_filenames)

        # Make sure skill name is unique.
        skill_name = f'{filename_no_extension}_retrieval_skill_{uuid4()}'
        retrieval_params = {
            'source': kb.name,
            'description': description,
        }
        file_retrieval_skill = self.skills.create(skill_name, 'retrieval', retrieval_params)
        agent.skills.append(file_retrieval_skill)
        self.update(agent.name, agent)

    def add_file(self, name: str, file_path: str, description: str, knowledge_base: str = None):
        """
        Add a file to the agent for retrieval.

        :param name: Name of the agent
        :param file_path: Path to the file to be added, or name of existing file.
        :param description: Description of the file. Used by agent to know when to do retrieval
        :param knowledge_base: Name of an existing knowledge base to be used. Will create a default knowledge base if not given.
        """
        self.add_files(name, [file_path], description, knowledge_base)

    def add_webpages(
            self,
            name: str,
            urls: List[str],
            description: str,
            knowledge_base: str = None,
            crawl_depth: int = 1,
            filters: List[str] = None
            ):
        """
        Add a list of webpages to the agent for retrieval.

        :param name: Name of the agent
        :param urls: List of URLs of the webpages to be added.
        :param description: Description of the webpages. Used by agent to know when to do retrieval.
        :param knowledge_base: Name of an existing knowledge base to be used. Will create a default knowledge base if not given.
        :param crawl_depth: How deep to crawl from each base URL. 0 = scrape given URLs only
        :param filters: Include only URLs that match these regex patterns
        """
        if not urls:
            return
        domain = ''
        path = ''
        agent = self.get(name)
        for url in urls:
            # Validate URLs.
            parsed_url = urlparse(url)
            domain = parsed_url.netloc.replace('.', '_')
            path = parsed_url.path.replace('/', '_')
        if knowledge_base is not None:
            kb = self.knowledge_bases.get(knowledge_base)
        else:
            kb_name = f'{name}_{domain}{path}_{uuid4()}_kb'
            kb = self._create_default_knowledge_base(agent, kb_name)

        # Insert crawled webpage.
        kb.insert_webpages(urls, crawl_depth=crawl_depth, filters=filters)

        # Make sure skill name is unique.
        skill_name = f'{domain}{path}_retrieval_skill_{uuid4()}'
        retrieval_params = {
            'source': kb.name,
            'description': description,
        }
        webpage_retrieval_skill = self.skills.create(skill_name, 'retrieval', retrieval_params)
        agent.skills.append(webpage_retrieval_skill)
        self.update(agent.name, agent)

    def add_webpage(
            self,
            name: str,
            url: str,
            description: str,
            knowledge_base: str = None,
            crawl_depth: int = 1,
            filters: List[str] = None):
        """
        Add a webpage to the agent for retrieval.

        :param name: Name of the agent
        :param file_path: URL of the webpage to be added, or name of existing webpage.
        :param description: Description of the webpage. Used by agent to know when to do retrieval.
        :param knowledge_base: Name of an existing knowledge base to be used. Will create a default knowledge base if not given.
        :param crawl_depth: How deep to crawl from each base URL. 0 = scrape given URLs only
        :param filters: Include only URLs that match these regex patterns
        """
        self.add_webpages(name, [url], description, knowledge_base=knowledge_base, crawl_depth=crawl_depth, filters=filters)

    def add_database(self, name: str, database: str, tables: List[str], description: str):
        """
        Add a database to the agent for retrieval.

        :param name: Name of the agent
        :param database: Name of the database to be added.
        :param tables: List of tables to be added.
        :param description: Description of the database. Used by agent to know when to do retrieval.
        """
        # Make sure database exists.
        db = self.databases.get(database)
        # Make sure tables exist.
        all_table_names = set([t.name for t in db.tables.list()])
        for t in tables:
            if t not in all_table_names:
                raise ValueError(f'Table {t} does not exist in database {database}.')

        # Make sure skill name is unique.
        skill_name = f'{database}_sql_skill_{uuid4()}'
        sql_params = {
            'database': database,
            'tables': tables,
            'description': description,
        }
        database_sql_skill = self.skills.create(skill_name, 'sql', sql_params)
        agent = self.get(name)

        if not agent.params:
            agent.params = {}
        if 'prompt_template' not in agent.params:
            # Set default prompt template. This is for langchain agent check.
            agent.params['prompt_template'] = 'using mindsdb sqltoolbox'

        agent.skills.append(database_sql_skill)
        self.update(agent.name, agent)

    def create(
            self,
            name: str,
            model: Union[Model, dict, str] = None,
            provider: str = None,
            skills: List[Union[Skill, str]] = None,
            params: dict = None,
            **kwargs) -> Agent:
        """
        Create new agent and return it

        :param name: Name of the agent to be created
        :param model: Model to be used by the agent
        :param skills: List of skills to be used by the agent. Currently only 'sql' is supported.
        :param params: Parameters for the agent

        :return: created agent object
        """
        skills = skills or []
        skill_names = []
        for skill in skills:
            if isinstance(skill, str):
                # Check if skill exists.
                # TODO what this line does?
                _ = self.skills.get(skill)
                skill_names.append(skill)
                continue
            # Create the skill if it doesn't exist.
            _ = self.skills.create(skill.name, skill.type, skill.params)
            skill_names.append(skill.name)

        if params is None:
            params = {}
        params.update(kwargs)

        if 'prompt_template' not in params:
            params['prompt_template'] = _DEFAULT_LLM_PROMPT

        if model is None:
            model = _DEFAULT_LLM_MODEL
        elif isinstance(model, Model):
            model = model.name
            provider = 'mindsdb'

        data = self.api.create_agent(self.project.name, name, model, provider, skill_names, params)
        return Agent.from_json(data, self)

    def update(self, name: str, updated_agent: Agent):
        """
        Update an agent by name.

        :param name: Name of the agent to be updated
        :param updated_agent: Agent with updated fields

        :return: updated agent object
        """
        updated_skills = set()
        for skill in updated_agent.skills:
            if isinstance(skill, str):
                # Skill must exist.
                _ = self.skills.get(skill)
                updated_skills.add(skill)
                continue
            try:
                # Create the skill if it doesn't exist.
                _ = self.skills.get(skill.name)
            except HTTPError as e:
                if e.response.status_code != 404:
                    raise e
                # Doesn't exist
                _ = self.skills.create(skill.name, skill.type, skill.params)
            updated_skills.add(skill.name)

        existing_agent = self.api.agent(self.project.name, name)
        existing_skills = set([s['name'] for s in existing_agent['skills']])
        skills_to_add = updated_skills.difference(existing_skills)
        skills_to_remove = existing_skills.difference(updated_skills)
        data = self.api.update_agent(
            self.project.name,
            name,
            updated_agent.name,
            updated_agent.model_name,
            list(skills_to_add),
            list(skills_to_remove),
            updated_agent.params
        )
        return Agent.from_json(data, self)

    def drop(self, name: str):
        """
        Drop an agent by name.

        :param name: Name of the agent to be dropped
        """
        _ = self.api.delete_agent(self.project.name, name)



from contextvars import ContextVar

context_storage = ContextVar('create_context')


def set_context(name: str, value: str):
    """
    Set context value to variable

    :param name: variable name
    :param value: variable value
    """
    data = context_storage.get({})
    data[name] = value

    context_storage.set(data)


def get_context(name: str) -> str:
    """
    Get context value fom variable

    :param name: variable name
    :return: variable value
    """

    data = context_storage.get({})
    return data.get(name)


def set_saving(name: str):
    """
    Set name of saving object to context

    :param name: namve of the object
    """
    set_context('saving', name)


def is_saving() -> bool:
    """
    Returns true if object is saved at the moment
    """

    return get_context('saving') is not None




from dataclasses import dataclass
from typing import List, Union

from mindsdb_sql.parser.ast import Show, Identifier
from mindsdb_sql.parser.dialects.mindsdb import CreateMLEngine, DropMLEngine

from mindsdb_sdk.utils.objects_collection import CollectionBase

from .handlers import Handler

@dataclass
class MLEngine:
    """
    :meta private:
    """
    name: str
    handler: str
    connection_data: dict


class MLEngines(CollectionBase):
    """

    **ML engines collection**

    Examples of usage:

    Get list

    >>> ml_engines = con.ml_engines.list()

    Get

    >>> openai_engine = con.ml_engines.openai1

    Create

    >>> con.ml_engines.create(
    ...    'openai1',
    ...    'openai',
    ...    connection_data={'api_key': '111'}
    ...)

    Drop

    >>>  con.ml_engines.drop('openai1')

    Upload BYOM model.
    After uploading a new ml engin will be availbe to create new model from it.

    >>> model_code = open('/path/to/model/code').read()
    >>> model_requirements = open('/path/to/model/requirements').read()
    >>> ml_engine = con.ml_engines.create_byom(
    ...    'my_byom_engine',
    ...    code=model_code,
    ...    requirements=model_requirements
    ...)

    """

    def __init__(self, api):
        self.api = api

    def list(self) -> List[MLEngine]:
        """
        Returns list of ml engines on server

        :return: list of ml engines
        """

        ast_query = Show(category='ml_engines')

        df = self.api.sql_query(ast_query.to_string())
        # columns to lower case
        cols_map = {i: i.lower() for i in df.columns}
        df = df.rename(columns=cols_map)

        return [
            MLEngine(**item)
            for item in df.to_dict('records')
        ]

    def get(self, name: str) -> MLEngine:
        """
        Get ml engine by name

        :param name
        :return: ml engine object
        """
        name = name.lower()
        for item in self.list():
            if item.name == name:
                return item
        raise AttributeError(f"MLEngine doesn't exist {name}")

    def create(self, name: str, handler: Union[str, Handler], connection_data: dict = None) -> MLEngine:
        """
        Create new ml engine and return it

        :param name: ml engine name, string
        :param handler: handler name, string or Handler
        :param connection_data: parameters for ml engine, dict, optional
        :return: created ml engine object
        """

        if isinstance(handler, Handler):
            handler = handler.name

        ast_query = CreateMLEngine(Identifier(name), handler, params=connection_data)

        self.api.sql_query(ast_query.to_string())

        return MLEngine(name, handler, connection_data)

    def create_byom(self, name: str, code: str, requirements: Union[str, List[str]] = None):
        """
        Create new BYOM ML engine and return it

        :param code: model python code in string
        :param requirements: requirements for model. Optional if there is no special requirements.
           It can be content of 'requirement.txt' file or list of strings (item for every requirement).
        :return: created BYOM ml engine object
        """

        if requirements is None:
            requirements = ''
        elif isinstance(requirements, list):
            requirements = '\n'.join(requirements)

        self.api.upload_byom(name, code, requirements)

        return MLEngine(name, 'byom', {})

    def drop(self, name: str):
        """
        Drop ml engine by name

        :param name: name
        """
        ast_query = DropMLEngine(Identifier(name))

        self.api.sql_query(ast_query.to_string())




from functools import wraps
from typing import List, Union
import io
import json

import requests
import pandas as pd
import validators

from mindsdb_sdk import __about__
from sseclient import SSEClient


def _try_relogin(fnc):
    @wraps(fnc)
    def wrapper(self, *args, **kwargs):
        try:
            return fnc(self, *args, **kwargs)
        except requests.HTTPError as e:
            if e.response.status_code != 401:
                raise e

            # try re-login
            try:
                self.login()
            except requests.HTTPError:
                raise e
            # call once more
            return fnc(self, *args, **kwargs)
    return wrapper


def _raise_for_status(response):
    # show response text in error
    if 400 <= response.status_code < 600:
        raise requests.HTTPError(f'{response.reason}: {response.text}', response=response)


class RestAPI:
    def __init__(self, url=None, login=None, password=None, api_key=None, is_managed=False, headers=None):

        self.url = url
        self.username = login
        self.password = password
        self.api_key = api_key
        self.is_managed = is_managed
        self.session = requests.Session()

        self.session.headers['User-Agent'] = f'python-sdk/{__about__.__version__}'
        if headers is not None:
            self.session.headers.update(headers)
        if self.api_key is not None:
            # Authenticate with API key instead of logging in, if present.
            self.session.headers['X-Api-Key'] = self.api_key
            return
        if login is not None:
            self.login()

    def login(self):
        managed_endpoint = '/api/login'
        cloud_endpoint = '/cloud/login'

        if self.is_managed:
            json = {'password': self.password, 'username': self.username}
            url = self.url + managed_endpoint
        else:
            json = {'password': self.password, 'email': self.username}
            url = self.url + cloud_endpoint
        r = self.session.post(url, json=json)

        # failback when is using managed instance with is_managed=False
        if r.status_code in (405, 404) and self.is_managed is False:
            # try managed instance login

            json = {'password': self.password, 'username': self.username}
            url = self.url + managed_endpoint
            r = self.session.post(url, json=json)

        _raise_for_status(r)

    @_try_relogin
    def sql_query(self, sql, database=None, lowercase_columns=False):

        if database is None:
            # it means the database is included in query
            database = 'mindsdb'
        url = self.url + '/api/sql/query'
        r = self.session.post(url, json={
            'query': sql,
            'context': {'db': database}
        })
        _raise_for_status(r)

        data = r.json()
        if data['type'] == 'table':
            columns = data['column_names']
            if lowercase_columns:
                columns = [i.lower() for i in columns]
            return pd.DataFrame(data['data'], columns=columns)
        if data['type'] == 'error':
            raise RuntimeError(data['error_message'])
        return None

    @_try_relogin
    def projects(self):
        # TODO not used yet

        r = self.session.get(self.url + '/api/projects')
        _raise_for_status(r)

        return pd.DataFrame(r.json())

    @_try_relogin
    def model_predict(self, project, model, data, params=None, version=None):
        data = data.to_dict('records')

        if version is not None:
            model = f'{model}.{version}'
        if params is None:
            params = {}
        url = self.url + f'/api/projects/{project}/models/{model}/predict'
        r = self.session.post(url, json={
            'data': data,
            'params': params
        })
        _raise_for_status(r)

        return pd.DataFrame(r.json())

    @_try_relogin
    def objects_tree(self, item=''):
        r = self.session.get(self.url + f'/api/tree/{item}')
        _raise_for_status(r)

        return pd.DataFrame(r.json())

    @staticmethod
    def read_file_as_bytes(file_path: str):
        """
        Read and return content of a file in bytes, given its path.
        :param file_path: Path of the file to read.
        :return: File content in bytes.
        """
        try:
            with open(file_path, 'rb+') as file:
                return file.read()
        except FileNotFoundError:
            raise Exception(f'File {file_path} does not exist.')
        except PermissionError:
            raise Exception(f'Permission denied when reading file {file_path}.')
        except Exception as e:
            raise Exception(f'Unknown error occurred when reading file {file_path} - {str(e)}')

    @staticmethod
    def read_dataframe_as_csv(data: pd.DataFrame):
        """
        Read and return content of a DataFrame as CSV in bytes.
        :param data: DataFrame to read.
        :return: DataFrame content as CSV in bytes.
        """
        fd = io.BytesIO()
        data.to_csv(fd, index=False)
        fd.seek(0)
        return fd.read()

    @staticmethod
    def read_file_as_webpage(url: str):
        """
        Read and return content of a file in bytes, given its URL.
        :param file_path: URL of the file to read.
        :return: File content in bytes.
        """
        data = requests.get(url)
        return data.content

    def upload_data(self, file_name: str, data: bytes):
        """
        Upload binary data to MindsDB.
        :param file_name: Name of the file.
        :param data: Binary data to upload.
        """
        # remove suffix from file if present
        name = file_name.split('.')[0]

        url = self.url + f'/api/files/{name}'
        r = self.session.put(
            url,
            data={
                'original_file_name':file_name,
                'name':name,
                'source_type':'file',
            },
            files={
                'file': (file_name, data)

            }
        )
        _raise_for_status(r)

    @_try_relogin
    def upload_file(self, name: str, data: Union[pd.DataFrame, str]):
        """
        Upload a file or a DataFrame to MindsDB.
        :param name: Name of the file or DataFrame.
        :param data: DataFrame data or file path.
        """
        if isinstance(data, pd.DataFrame):
            data_in_bytes = self.read_dataframe_as_csv(data)
        elif validators.url(data):
            data_in_bytes = self.read_file_as_webpage(data)
        else:
            data_in_bytes = self.read_file_as_bytes(data)

        self.upload_data(name, data_in_bytes)

    @_try_relogin
    def get_file_metadata(self, name: str) -> dict:
        # No endpoint currently to get single file.
        url = self.url + f'/api/files/'
        r = self.session.get(url)
        _raise_for_status(r)
        all_file_metadata = r.json()
        for metadata in all_file_metadata:
            if metadata.get('name', None) == name:
                return metadata
        r.status_code = 404
        raise requests.HTTPError(f'Not found: No file named {name} found', response=r)

    @_try_relogin
    def upload_byom(self, name: str, code: str, requirements: str):

        url = self.url + f'/api/handlers/byom/{name}'
        r = self.session.put(
            url,
            files={
                'code': code,
                'modules': requirements,
            }
        )
        _raise_for_status(r)

    def status(self) -> dict:

        r = self.session.get(self.url + '/api/status')
        _raise_for_status(r)

        return r.json()

    # TODO: Different endpoints should be refactored into their own classes.
    #
    # Agents operations.
    @_try_relogin
    def agents(self, project: str):
        r = self.session.get(self.url + f'/api/projects/{project}/agents')
        _raise_for_status(r)

        return r.json()

    @_try_relogin
    def agent(self, project: str, name: str):
        r = self.session.get(self.url + f'/api/projects/{project}/agents/{name}')
        _raise_for_status(r)

        return r.json()

    @_try_relogin
    def agent_completion(self, project: str, name: str, messages: List[dict]):
        url = self.url + f'/api/projects/{project}/agents/{name}/completions'
        r = self.session.post(
            url,
            json={
                'messages': messages
            }
        )
        _raise_for_status(r)

        return r.json()

    @_try_relogin
    def agent_completion_stream(self, project: str, name: str, messages: List[dict]):
        url = self.url + f'/api/projects/{project}/agents/{name}/completions/stream'
        stream = self.session.post(url, json={'messages': messages}, stream=True)
        client = SSEClient(stream)
        for chunk in client.events():
            # Stream objects loaded from SSE events 'data' param.
            yield json.loads(chunk.data)

    @_try_relogin
    def create_agent(self, project: str, name: str, model: str = None, provider: str = None, skills: List[str] = None, params: dict = None):
        url = self.url + f'/api/projects/{project}/agents'
        r = self.session.post(
            url,
            json={
                'agent': {
                    'name': name,
                    'model_name': model,
                    'provider': provider,
                    'skills': skills,
                    'params': params
                }
            }
        )
        _raise_for_status(r)
        return r.json()

    @_try_relogin
    def update_agent(
            self,
            project: str,
            name: str,
            updated_name: str,
            updated_model: str,
            skills_to_add: List[str],
            skills_to_remove: List[str],
            updated_params: dict
            ):
        url = self.url + f'/api/projects/{project}/agents/{name}'
        r = self.session.put(
            url,
            json={
                'agent': {
                    'name': updated_name,
                    'model_name': updated_model,
                    'skills_to_add': skills_to_add,
                    'skills_to_remove': skills_to_remove,
                    'params': updated_params
                }
            }
        )
        _raise_for_status(r)
        return r.json()

    @_try_relogin
    def delete_agent(self, project: str, name: str):
        url = self.url + f'/api/projects/{project}/agents/{name}'
        r = self.session.delete(url)
        _raise_for_status(r)

    # Skills operations.
    @_try_relogin
    def skills(self, project: str):
        r = self.session.get(self.url + f'/api/projects/{project}/skills')
        _raise_for_status(r)

        return r.json()

    @_try_relogin
    def skill(self, project: str, name: str):
        r = self.session.get(self.url + f'/api/projects/{project}/skills/{name}')
        _raise_for_status(r)

        return r.json()

    @_try_relogin
    def create_skill(self, project: str, name: str, type: str, params: dict):
        url = self.url + f'/api/projects/{project}/skills'
        r = self.session.post(
            url,
            json={
               'skill': {
                    'name': name,
                    'type': type,
                    'params': params
                }
            }
        )
        _raise_for_status(r)

        return r.json()

    @_try_relogin
    def update_skill(
            self,
            project: str,
            name: str,
            updated_name: str,
            updated_type: str,
            updated_params: dict
            ):
        url = self.url + f'/api/projects/{project}/skills/{name}'
        r = self.session.put(
            url,
            json={
               'skill': {
                    'name': updated_name,
                    'type': updated_type,
                    'params': updated_params
                }
            }
        )
        _raise_for_status(r)
        return r.json()

    @_try_relogin
    def delete_skill(self, project: str, name: str):
        url = self.url + f'/api/projects/{project}/skills/{name}'
        r = self.session.delete(url)
        _raise_for_status(r)

    # Knowledge Base operations.
    @_try_relogin
    def insert_files_into_knowledge_base(self, project: str, knowledge_base_name: str, file_names: List[str]):
        r = self.session.put(
            self.url + f'/api/projects/{project}/knowledge_bases/{knowledge_base_name}',
            json={
                'knowledge_base': {
                    'files': file_names
                }
            }
        )
        _raise_for_status(r)

        return r.json()

    @_try_relogin
    def insert_webpages_into_knowledge_base(self, project: str, knowledge_base_name: str, urls: List[str], crawl_depth: int = 1, filters: List[str] = None):
        data = {
            'urls': urls,
            'crawl_depth': crawl_depth,
            'filters': [] if filters is None else filters
        }
        r = self.session.put(
            self.url + f'/api/projects/{project}/knowledge_bases/{knowledge_base_name}',
            json={
                'knowledge_base': data
            }
        )
        _raise_for_status(r)

        return r.json()
